from deap import benchmarks
import math
import numpy as np

# Some test objective functions to help guide optimizer choices
# -------------------------------------------------------------
#
# We'll use DEAP's set of groovy benchmarks.
# Your mileage may vary
# See pretty pictures at https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks
# Once can easily add more from DEAP or elsewhere

## Basis of tricky functions


def schwefel_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schwefel
    u_squished = [ 1000*(ui**1.1-0.5) for ui in u ]
    return 0.001*benchmarks.schwefel(u_squished)[0]


def griewank_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.griewank
    u_squished = [ 1200*(ui**1.1-0.5) for ui in u ]
    return benchmarks.griewank(u_squished)[0]


def rastrigin_on_cube(u:[float])->float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.rastrigin
    u_squished = [ 10.24*(ui**1.1-0.5) for ui in u ]
    return 0.01*benchmarks.rastrigin(u_squished)[0]

## Combinations

def deap_on_cube(u:[float])->float:
    return schwefel_on_cube(u) + griewank_on_cube(u) + rastrigin_on_cube(u)


# According to http://infinity77.net/global_optimization/test_functions.html#test-functions-index
# there are some really hard ones
# See https://github.com/andyfaff/ampgo/blob/master/%20ampgo%20--username%20andrea.gavana%40gmail.com/go_benchmark.py


def damavandi_on_cube(u:[float])->float:
    """ A multi-dimensional extension of Damavandi's function """
    u0 = u[0]
    dama = [ damavandi2(u0,u_) for u_ in u[1:] ]
    return float(np.sum(dama))


def damavandi2(u1,u2)->float:
    """ Pretty evil function this one """
    # http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.Damavandi
    x1 = u1/14.
    x2 = u2/14.
    numerator = math.sin(math.pi*(x1 - 2.0))*math.sin(math.pi*(x2 - 2.0))
    denumerator = (math.pi**2)*(x1 - 2.0)*(x2 - 2.0)
    factor1 = 1.0 - (abs(numerator / denumerator))**5.0
    factor2 = 2 + (x1 - 7.0)**2.0 + 2*(x2 - 7.0)**2.0
    return factor1*factor2


def paviani_on_cube(u:[float])->float:
    # http://infinity77.net/global_optimization/test_functions_nd_P.html#go_benchmark.Paviani
    x = np.array([ 2.001+(ui**1.1)*6.996 for ui in u])
    return float( np.sum(np.log(x-2)**2.0 + np.log(10.0 - x)**2.0) - np.prod(x)**0.2 )


AN_OBJECTIVE = deap_on_cube
OBJECTIVES   = [schwefel_on_cube, rastrigin_on_cube, griewank_on_cube, deap_on_cube,
                paviani_on_cube, damavandi_on_cube ]


if __name__=="__main__":
    for objective in OBJECTIVES:
        objective(u=[0.5,0.5,0.5])


from pprint import pprint
from timemachines.optimizers.compendium import OPTIMIZERS
from timemachines.optimizers.objectives import OBJECTIVES


def optimizer_name(solver):
    return solver.__name__.replace('_cube','').replace('_skew',' (skew) ').replace('_scaled',' (scaled) ')

def debug():
    for n_trials in [2,10,40]:
        for n_dim in [2,3]:
            for optimizer in OPTIMIZERS:
                for objective in OBJECTIVES:
                    try:
                        optimizer(objective, n_trials=n_trials,n_dim=n_dim,with_count=True)
                    except Exception as e:
                        print(e)
                        print(optimizer_name(optimizer) + ' fails on ' + objective.__name__ +
                              ' in dimension '+str(n_dim)+' when n_trials='+str(n_trials))


def comparison():
    n_trials,n_dim = 300, 5
    comparison = sorted([(objective.__name__,optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True),
                          optimizer_name(optimizer))
                         for optimizer in OPTIMIZERS for objective in OBJECTIVES])
    pprint(comparison)


if __name__=='__main__':
    # Beware that some optimizers may take more function evals than instructed
    # Others may multi-thread
    # This is here more as a quick unit test than a legitimate "comparison"
    comparison()








from timemachines.optimizers.hyperoptcube import hyperopt_cube
from timemachines.optimizers.shgocube import shgo_cube
from timemachines.optimizers.optunacube import optuna_cube
from timemachines.optimizers.pysotcube import pysot_cube
from timemachines.optimizers.powellcube import powell_cube
from timemachines.optimizers.axcube import ax_cube
from timemachines.optimizers.platypuscube import genetic_cube, evolutionary_cube
from timemachines.optimizers.pymoocube import nelder_cube, ctaea_cube, nsga3_cube, pattern_cube

OPTIMIZERS = [ ax_cube, hyperopt_cube, shgo_cube, optuna_cube, pysot_cube,
               powell_cube, evolutionary_cube, genetic_cube, nelder_cube,
                nsga3_cube, pattern_cube ]

FAILING_OPTIMIZERS = [ ctaea_cube ]

if __name__=='__main__':
    from timemachines.optimizers.objectives import AN_OBJECTIVE
    minima = [ optimizer(AN_OBJECTIVE, n_trials=50, n_dim=5) for optimizer in OPTIMIZERS]


from typing import List
import numpy as np
import divinity as dv
from timemachines.conventions import to_int_log_space, Y_TYPE, K_TYPE
from timemachines.utilities import suppression
from timemachines.utilities.suppression import suppress_output
from timemachines.plotting import prior_plot

# I don't really recommend this guy. For one thing the logs are hard to suppress ;)


def divinity_seasonal(y:Y_TYPE, s, k:K_TYPE, a=None, t=None, e=None, r=0.5):
    """ A partial wrapping of the divinity library with notable limitations:

         - Fits every invocation (slow)
         - Ignores exogenous variables

        This guy's a tad noisy for my liking.
    """

    # Allow r to control buffer length and burn-in
    max_buffer_len, burnin = to_int_log_space(p=r, bounds=[(250, 5000), (120, 140)])
    assert max_buffer_len>100, 'divine library assumes 100 lags at least '

    if s is None:
        s = dict(buffer=[])

    if y is None:
        return None, s, None   # Ignore suggestion to fit offline

    if isinstance(y,List):
        y = y[0]         # Ignore exogenous variables

    # Update buffer
    s['buffer'].append(y)
    if len(s['buffer'])>max_buffer_len+1000:
        s['buffer'] = s['buffer'][-max_buffer_len:]

    # Fit and predict, if warm, or just last value
    if len(s['buffer']) < burnin:
        return y, s, None
    else:
        with suppress_output():  # this doesn't work :(
            model = dv.divinity(forecast_length=k,optimise_trend_season_features = True)
            model.fit(np.array(s['buffer']))
            x = model.predict()[k-1]
        return x, s, None


if __name__=='__main__':
    err = prior_plot(f=divinity_seasonal, k=1, n=150)
    pass




from timemachines.optimization import optimize
from timemachines.synthetic import brownian_with_exogenous
from timemachines.optimizers.alloptimizers import OPTIMIZERS
from timemachines.skaters.allskaters import SKATERS
import time
import random


def optimize_random_skater():
    start_time = time.time()
    optimizer = random.choice(OPTIMIZERS)
    f = random.choice(SKATERS)
    print('Running '+str(optimizer.__name__))
    r_star = optimize(f=f, ys=brownian_with_exogenous(n=150),
                      n_trials=15, optimizer=optimizer)
    print("Best hyper-param is " + str(r_star))
    print('Took ' + str( (time.time()-start_time)/60 ) + ' minutes.' )



if __name__=='__main__':
    optimize_random_skater()


from ax import optimize
from logging import CRITICAL
from ax.utils.common.logger import get_logger
rt = get_logger('ax')
rt.setLevel(CRITICAL)
import warnings
warnings.filterwarnings("ignore")
from funcy import print_durations
from timemachines.optimizers.objectives import OBJECTIVES


def ax_cube(objective, n_trials, n_dim, with_count=False):
    global feval_count
    feval_count = 0

    def evaluation_func(prms):
        global feval_count
        feval_count += 1
        return objective([prms["u" + str(i)] for i in range(n_dim)])

    parameters = [{
        "name": "u" + str(i),
        "type": "range",
        "bounds": [0.0, 1.0],
    } for i in range(n_dim)]
    best_parameters, best_values, experiment, model = optimize(parameters=parameters,
                                                               evaluation_function=evaluation_func,
                                                               minimize=True,
                                                               total_trials=n_trials)
    return (best_values[0]['objective'], feval_count) if with_count else best_values[0]['objective']


@print_durations()
def demo():
    for objective in OBJECTIVES:
        print(ax_cube(objective, n_trials=50, n_dim=2, with_count=True))

if __name__ == '__main__':
    demo()




from timemachines.optimization import optimize
from timemachines.synthetic import brownian_with_exogenous
from timemachines.optimizers.compendium import OPTIMIZERS
from timemachines.skaters.pmd import pmd_hyperparams, pmd_auto
from pprint import pprint
import random


def optimize_dlm():
    optimizer = random.choice(OPTIMIZERS)
    print('Running '+str(optimizer.__name__))
    r_star = optimize(f=pmd_auto, ys=brownian_with_exogenous(n=300),
                      n_trials=25, optimizer=optimizer)
    s = pmd_hyperparams(s=dict(), r=r_star)
    print("Best hyper-param is " + str(r_star)+' with interpretation ')
    pprint(s)


from timemachines.stochastictests.optimizedlm import optimize_dlm
from timemachines.stochastictests.optimizerandomskater import optimize_random_skater
from timemachines.stochastictests.rundlm import run_dlm

STOCHASTIC_TESTS = [ run_dlm, optimize_dlm, optimize_random_skater ]


from timemachines.stochastictests.compendium import STOCHASTIC_TESTS
import time
import random
TIMEOUT = 60*5

# Regression tests run occasionally to check various parts of hyper-param spaces, etc.

if __name__=='__main__':
    start_time = time.time()
    elapsed = time.time()-start_time
    while elapsed < TIMEOUT:
        a_test = random.choice(STOCHASTIC_TESTS)
        print('Running '+str(a_test.__name__))
        a_test()
        elapsed = time.time() - start_time



from platypus import NSGAII, Problem, Real, EvolutionaryStrategy, GeneticAlgorithm


def evolutionary_cube(objective,n_trials,n_dim,with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials,n_dim=n_dim, with_count=with_count,strategy=EvolutionaryStrategy)


def genetic_cube(objective,n_trials,n_dim,with_count=False):
    return platypus_cube(objective=objective, n_trials=n_trials,n_dim=n_dim,with_count=with_count,strategy=GeneticAlgorithm)


def platypus_cube(objective, n_trials, n_dim, strategy, with_count=False):

    global feval_count
    feval_count = 0

    def _objective(vars):
        global feval_count
        feval_count += 1
        return objective(list(vars))

    problem = Problem(n_dim, 1, 0)
    problem.types[:] = [Real(0.0, 1.0)]*n_dim
    problem.constraints[:] = "<=0"
    problem.function = _objective

    algorithm = strategy(problem)
    algorithm.run(n_trials)
    feasible_solution_obj = [s.objectives[0] for s in algorithm.result if s.feasible]
    best_obj = min(feasible_solution_obj)
    return (best_obj, feval_count) if with_count else best_obj


if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES
    for objective in OBJECTIVES:
        print(evolutionary_cube(objective, n_trials=50, n_dim=6, with_count=True))
        print(genetic_cube(objective, n_trials=50, n_dim=6, with_count=True))



from pymoo.optimize import minimize
from pymoo.model.problem import Problem
import numpy as np
from pymoo.factory import get_algorithm, get_termination, get_reference_directions

REF_DIRS = get_reference_directions("das-dennis", 1, n_partitions=6)


def brkga_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="brkga", n_dim=n_dim, with_count=with_count)


def nelder_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nelder-mead", n_dim=n_dim, with_count=with_count)


def cmaes_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="cmaes", n_dim=n_dim, with_count=with_count)


def nsga2_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def rnsga2_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def rnsga3_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def unsga3_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="unsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def moead_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="moead", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def pattern_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="pattern-search", n_dim=n_dim, with_count=with_count)


def ctaea_cube(objective, n_trials,n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="ctaea", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def nsga3_cube(objective,  n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


def de_cube(objective, n_trials, n_dim, with_count=False):
    return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", n_dim=n_dim, with_count=with_count)


def pymoo_cube(objective, n_trials, method_name, n_dim, with_count, ref_dirs=None):

    class ObjectiveProblem(Problem):

        def __init__(self):
            super().__init__(n_var=n_dim, n_obj=1, n_constr=0, xl=0.0, xu=1.0)
            self.feval_count = 0

        def _evaluate(self, x, out, *args, **kwargs):
            """ vectorized  """
            self.feval_count = self.feval_count + len(x)
            out["F"] = np.array([objective(u) for u in x])

    try:
        algorithm = get_algorithm(method_name, ref_dirs=ref_dirs)
    except ValueError:
        algorithm = get_algorithm(method_name)
    termination = get_termination("n_eval", n_trials)
    problem = ObjectiveProblem()

    result = minimize(problem=problem,
                      algorithm=algorithm,
                      termination=termination,
                      seed=None,
                      verbose=False,
                      display=None,
                      callback=None,
                      return_least_infeasible=False,
                      save_history=False
                      )
    f_min = result.F[0]
    return (f_min, problem.feval_count) if with_count else f_min


PYMOO_BROKEN = [de_cube,rnsga2_cube]
PYMOO_GOOD = [nelder_cube,ctaea_cube,nsga3_cube,pattern_cube]
PYMOO_BAD = [brkga_cube,nsga2_cube]

if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES
    for solver in PYMOO_BAD:
        print(' ')
        for objective in OBJECTIVES:
            print((objective.__name__,solver(objective, n_trials=100, n_dim=3, with_count=True)))
    print('done')


import optuna
from optuna.logging import CRITICAL
from timemachines.optimizers.objectives import OBJECTIVES


def optuna_cube(objective, n_trials,n_dim, with_count=False):

    global feval_count
    feval_count = 0

    def cube_objective(trial):
        global feval_count
        us = [ trial.suggest_float('u'+str(i),0,1) for i in range(n_dim)]
        feval_count += 1
        return objective(us)

    optuna.logging.set_verbosity(CRITICAL)
    study = optuna.create_study()
    study.optimize(cube_objective,n_trials=n_trials)
    if with_count:
        return study.best_value, feval_count
    else:
        return study.best_value


if __name__=='__main__':
    for objective in OBJECTIVES:
        print((objective.__name__,optuna_cube(objective, n_trials=100, n_dim=6, with_count=True)))



import pmdarima as pm
from timemachines.conventions import to_int_log_space, initialize_buffers, separate_observations, update_buffers, \
    Y_TYPE, A_TYPE, R_TYPE
from timemachines.plotting import prior_plot_exogenous


# pmd-arima skaters


def pmd_hyperparams(s, r:R_TYPE):
    # One way to interpret hyper-params r for pmd_arima models
    p_bounds, q_bounds, n_fit_bounds = (1, 10), (1, 10), (1, 500)
    s['n_burn'] = 50
    s['alpha'] = 0.25  # Defines confidence interval
    s['buffer_len'] = 5000
    s['max_p'], s['max_q'], s['n_fit'] = to_int_log_space(r, bounds=[p_bounds, q_bounds, n_fit_bounds])
    return s


def pmd_or_last_value(s, k:int, exog:[float], y0:float):
    # Predict k steps ahead, or fall back to last value
    if s.get('model') is None:
        return y0, s, None
    else:
        x, ntvl = s['model'].predict(n_periods=k, X=exog, return_conf_int=True, alpha=s['alpha'])
        w = ntvl[k - 1][1] - ntvl[k - 1][0]
        return x, s, w


def pmd_auto(y, s, k, a, t, e, r):
    """ One way to use pmdarima auto_arima

            - Contemporaneous y[1:] variables are used as exogenous 'X' in pmdarima
            - This only works for k=1

        :returns: x, s', w
    """
    if s is None:
        s = dict()
        s = pmd_hyperparams(s=s, r=r)
        s = initialize_buffers(s=s,y=y)

    if y is not None:
        assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
        y0, exog = separate_observations(y=y, dim=s['dim'])
        s = update_buffers(s=s,a=a,exog=exog,y0=y0)
        s['staleness'] += 1
        stale = (len(s['buffer']) >= s['n_burn']) and (s['staleness'] >= s['n_fit'])
        first_time = len(s['buffer']) == s['n_burn']
        if first_time or stale:
            none_, s, _ = pmd_auto(y=None, s=s, k=k, a=a, t=t, e=e, r=r)  # Fit the model
            assert none_ is None
        elif s['model'] is not None:
            s['model'].update([y0], X=exog)
        return pmd_or_last_value(s=s, k=k, exog=exog, y0=y0)

    if y is None:
        # Fitting the model
        X = s.get('exogenous') or None
        s['model'] = pm.auto_arima(y=s['buffer'], X=X, start_p=1, start_q=1, start_P=1, start_Q=1,
                                   max_p=s['max_p'], max_q=s['max_q'], max_P=5, max_Q=5, seasonal=True,
                                   stepwise=True, suppress_warnings=True, D=10, max_D=10,
                                   error_action='ignore')
        s['staleness'] = 0
        return None, s, None


if __name__ == '__main__':
    err = prior_plot_exogenous(f=pmd_auto, k=1, n=500, r=0.05)
    pass



from pydlm import dlm, trend, seasonality
from timemachines.conventions import dimension, to_log_space, separate_observations, update_buffers, \
    Y_TYPE, A_TYPE, R_TYPE
from timemachines.plotting import prior_plot_exogenous
import math
import numpy as np
import matplotlib.pyplot as plt

# pydlm skaters
# TODO:
# TODO: k-step ahead prediction


def dlm_seasonal_hyperparams(s, r:R_TYPE):
    # One way to interpret hyper-params r for pmd_arima models
    # e.g. dlm([]) + trend(degree=2, 0.98) + + seasonality(period = 7, 0.98)
    degree_bounds = (0.1,4.1)
    initial_discount = (0.90,0.97)  # Tuned subsequently so doesn't matter that much
    seasonality_choices = [0,7,24,3*24,4*24]
    seasonality_choice_bounds = [0.2,4.4]
    bounds=[seasonality_choice_bounds, degree_bounds, initial_discount ]
    _period_choice, _degree, s['discount'] = to_log_space(r, bounds=bounds)
    period_choice = int(_period_choice)
    assert 0<= period_choice <= 3
    s['degree'], s['period'] = int(_degree), seasonality_choices[period_choice]
    s['n_burn'] = max(50,2*s['period'])
    s['n_fit'] = 100  # How often to tune discount
    return s


def dlm_seasonal_or_last_value(s, k:int, y:Y_TYPE):
    # Predict using model, or fall back to last value
    num_obs = len(s['model'].data) if s.get('model') else 0
    if num_obs < s['n_burn']:
        y0, _ = separate_observations(y,s['dim'])
        return y0, s, None
    else:
        assert k==1,'only k==1 for now' # TODO: Fix
        x_mean, x_var = s['model'].predict()
        x = x_mean[0, 0]
        v = x_var[0,0]
        try:
            w = math.sqrt(v)
        except:
            print('Warning - negative var returned '+str(v))
            w = math.sqrt(abs(v))
        return x, s, w


def dlm_seasonal(y, s, k, a, t, e, r):
    """ One way to use dlm
        :returns: x, s', w
    """
    if s is None:
        s = dict()
        s = dlm_seasonal_hyperparams(s=s, r=r)
        s['dim'] = dimension(y)
        if s['period']>0:
            s['model'] = dlm([]) + trend(s['degree'], s['discount']) + seasonality(s['period'], s['discount'])
        else:
            s['model'] = dlm([]) + trend(s['degree'], s['discount'])

    if y is not None:
        assert isinstance(y, float) or len(y) == s['dim'], ' Cannot change dimension of input in flight '
        y0, exog = separate_observations(y=y,dim=s['dim'])
        y0_passed_in = None if np.isnan(y0) else None  # pydlm uses None for missing values
        s['model'].append([y0])
        num_obs = len(s['model'].data) if s.get('model') else 0
        if num_obs % s['n_fit'] == s['n_fit']-1:
            _, s, _ = dlm_seasonal(y=None,s=s,k=k,a=a,t=t,e=10,r=r) # Fit
        s['model'].fitForwardFilter()
        s['model'].fitBackwardSmoother()
        return dlm_seasonal_or_last_value(s=s, k=k, y=y)

    if y is None:
        s['model'].tune() # Tunes discount factors
        s['model'].fit()
        return None, s, None


if __name__ == '__main__':
    err = prior_plot_exogenous(f=dlm_seasonal, k=1, n=200, r=0.05)
    plt.figure()
    print('done')
    pass



from hyperopt import fmin, hp, tpe, Trials


def hyperopt_cube(objective, n_trials, n_dim, with_count=False):
    """ Minimize a function on the cube using HyperOpt, and audit # of function calls
       :param objective:    function on (0,1)^n_dim
       :param n_trials:     Guideline for function evaluations
       :param n_dim:
       :param with_count:
       :return:
    """
    hp_space = dict( [('u'+str(i), hp.uniform('u'+str(i), 0, 1)) for i in range(n_dim)])

    global feval_count
    feval_count = 0

    def _objective(hps):
        global feval_count
        feval_count += 1
        us = [ hps['u'+str(i)] for i in range(n_dim)]
        return objective(us)

    trls = Trials()
    res = fmin(_objective, space=hp_space, algo=tpe.suggest, trials=trls, max_evals=n_trials, show_progressbar=False)
    return (trls.best_trial['result']['loss'], feval_count) if with_count else trls.best_trial['result']['loss']





from timemachines.skaters.dlm import dlm_seasonal
from timemachines.evaluation import quick_brown_fox_randomized


def run_dlm():
    err = quick_brown_fox_randomized(f=dlm_seasonal, n=1000)


if __name__=='__main__':
    run_dlm()


from scipy.optimize import minimize

global feval_count
feval_count=0


def powell_cube(objective, n_trials, n_dim, with_count=False):
    bounds = [(0,1) ]*n_dim

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count +=1
        return objective(list(x))

    result = minimize(_objective, x0=[0]*n_dim, method='powell',bounds=bounds, options={'maxfev':n_trials,'maxiter':n_trials})
    return (_objective(result.x), feval_count) if with_count else _objective(result.x)


if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES
    for objective in OBJECTIVES:
        print(powell_cube(objective, n_trials=5, n_dim=50, with_count=True))



from poap.controller import BasicWorkerThread, ThreadController
from pySOT.experimental_design import SymmetricLatinHypercube
from pySOT.strategy import SRBFStrategy
from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant
from pySOT.optimization_problems.optimization_problem import OptimizationProblem
import numpy as np


class GenericProblem(OptimizationProblem):

    def __init__(self, dim, objective):
        self.dim = dim
        self.min = 0
        self.minimum = np.zeros(dim)
        self.lb = 0 * np.ones(dim)
        self.ub = 1 * np.ones(dim)
        self.int_var = np.array([])
        self.cont_var = np.arange(0, dim)
        self.objective = objective
        self.info = str(dim) + "-dimensional objective function " + objective.__name__
        self.feval_count = 0

    def eval(self, x):
        """Evaluate the objective x

        :param x: Data point
        :type x: numpy.array
        :return: Value at x
        :rtype: float
        """
        self.__check_input__(x)
        d = float(self.dim)
        self.feval_count = self.feval_count+1
        return self.objective(list(x))


def pysot_cube( objective, n_trials, n_dim, with_count=False):
    """ Minimize
    :param objective:
    :param n_trials:
    :param n_dim:
    :param with_count:
    :return:
    """

    num_threads = 1
    max_evals = n_trials
    gp = GenericProblem(dim=n_dim, objective=objective)
    rbf = RBFInterpolant(dim=n_dim, lb=np.array([0.0]*n_dim), ub=np.array([1.0]*n_dim), kernel=CubicKernel(), tail=LinearTail(n_dim))
    slhd = SymmetricLatinHypercube(dim=n_dim, num_pts=2 * (n_dim + 1))

    # Create a strategy and a controller
    controller = ThreadController()
    controller.strategy = SRBFStrategy(
        max_evals=max_evals, opt_prob=gp, exp_design=slhd, surrogate=rbf, asynchronous=True
    )

    # Launch the threads and give them access to the objective function
    for _ in range(num_threads):
        worker = BasicWorkerThread(controller, gp.eval)
        controller.launch_worker(worker)

    # Run the optimization strategy
    result = controller.run()
    return (result.value, gp.feval_count) if with_count else result.value


if __name__ == '__main__':
    from timemachines.optimizers.objectives import AN_OBJECTIVE
    print(pysot_cube(AN_OBJECTIVE, n_trials=100, n_dim=3, with_count=True))


from scipy.optimize import shgo


def shgo_cube(objective, n_trials, n_dim, with_count: bool = False):
    """ Minimize a function on the cube using SHGO
    :param objective:    function on (0,1)^n_dim
    :param n_trials:
    :param n_dim:
    :param with_count:
    :return:
    """
    bounds = [(0, 1)] * n_dim

    global feval_count
    feval_count = 0

    def _objective(x):
        global feval_count
        feval_count += 1
        return objective(list(x))

    # Try to induce roughly the right number of function evaluations
    n_trials_reduced = n_trials - 20
    n_iters = int(1+n_trials/40)
    n = int(5+n_trials/20)
    result = shgo(_objective, bounds, n=n, iters=n_iters, options={'maxfev': n_trials_reduced,
                                               'minimize_every_iter': False,
                                               'maxfun': n_trials_reduced,
                                               'minimizer_kwargs':{'maxiter':10}},
                                sampling_method='sobol')
    return (result.fun, feval_count) if with_count else result.fun


if __name__ == '__main__':
    from timemachines.optimizers.objectives import OBJECTIVES

    for objective in OBJECTIVES:
        print(shgo_cube(objective, n_trials=50, n_dim=5, with_count=True))
